from jaseci.utils.test_core import CoreTest


class ObjectApiTest(CoreTest):
    """Unit tests for Jac Walker APIs"""

    fixture_src = __file__

    def test_object_set(self):
        ret = self.call(
            self.mast,
            [
                "object_set",
                {"obj": self.mast.jid, "ctx": {"name": "MYNEWNAME", "jid": "MYNEWJID"}},
            ],
        )
        self.assertEqual(ret["name"], "MYNEWNAME")
        self.assertNotEqual(ret["jid"], "MYNEWJID")
        self.assertEqual(self.mast.name, "MYNEWNAME")
        self.assertNotEqual(self.mast.jid, "MYNEWJID")



"""
Walker interpreter for jac code in AST form

This interpreter should be inhereted from the class that manages state
referenced through self.
"""
from jaseci.prim.node import Node
from jaseci.jac.interpreter.interp import Interp
from jaseci.jac.jac_set import JacSet
from jaseci.jac.machine.jac_scope import JacScope
from jaseci.jac.ir.jac_code import jac_ir_to_ast


class WalkerInterp(Interp):
    """Jac interpreter mixin for objects that will execute Jac code"""

    # Walker only executes statements, sentinels handle attr_stmts

    def run_walker(self, jac_ast):
        """
        walker: KW_ASYNC? KW_WALKER NAME namespaces? walker_block;
        """
        kid = self.set_cur_ast(jac_ast)
        self.scope_and_run(
            jac_ast if jac_ast.name == "walker_block" else kid[-1],
            self.run_walker_block,
            scope_name=f"w_run:{jac_ast.loc_str()}",
        )

    def run_walker_block(self, jac_ast):
        """
        walker_block:
            LBRACE attr_stmt* walk_entry_block? (
                statement
                | walk_activity_block
            )* walk_exit_block? RBRACE;
        """
        kid = self.set_cur_ast(jac_ast)
        act_list = self.current_node.get_architype().get_entry_abilities()
        self.auto_trigger_node_actions(act_list=act_list)

        for i in kid:
            if i.name == "walk_entry_block":
                self.run_walk_entry_block(i)
            if i.name == "statement":
                self.run_statement(i)
            if i.name == "walk_activity_block":
                self.run_walk_activity_block(i)

        act_list = self.current_node.get_architype().get_exit_abilities()
        self.auto_trigger_node_actions(act_list=act_list)

        if not self.yielded and kid[-2].name == "walk_exit_block":
            self.run_walk_exit_block(kid[-2])

    def run_node_ctx_block(self, jac_ast):
        """
        node_ctx_block: name_list code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        for i in self.run_name_list(kid[0]):
            if self.current_node.get_architype().is_instance(i):
                self.run_code_block(kid[1])
                return

    def run_walk_entry_block(self, jac_ast):
        """
        walk_entry_block: KW_WITH KW_ENTRY code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        if self.current_step == 0:
            self.in_entry_exit = True
            self.run_code_block(kid[2])
            self.in_entry_exit = False

    def run_walk_exit_block(self, jac_ast):
        """
        walk_exit_block: KW_WITH KW_EXIT code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        self._stopped = None
        if len(self.next_node_ids) == 0:
            self.in_entry_exit = True
            self.run_code_block(kid[2])
            self.in_entry_exit = False

    def run_walk_activity_block(self, jac_ast):
        """
        walk_activity_block: KW_WITH KW_ACTIVITY code_block;
        """
        kid = self.set_cur_ast(jac_ast)
        self.run_code_block(kid[2])

    def run_walker_action(self, jac_ast):
        """
        walker_action:
            ignore_action
            | take_action
            | disengage_action
            | yield_action;
        """
        kid = self.set_cur_ast(jac_ast)
        expr_func = getattr(self, f"run_{kid[0].name}")
        expr_func(kid[0])

    def run_ignore_action(self, jac_ast):
        """
        ignore_action: KW_IGNORE expression SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        self.run_expression(kid[1])
        result = self.pop().value
        if isinstance(result, Node):
            self.ignore_node_ids.add_obj(result)
        elif isinstance(result, JacSet):
            self.ignore_node_ids.add_obj_list(result)
        else:
            self.rt_error(f"{result} is not ignorable type (i.e., nodes)", kid[1])

    def run_take_action(self, jac_ast):
        """
        take_action:
            KW_TAKE (COLON NAME)? expression (SEMI | else_stmt);
        """
        kid = self.set_cur_ast(jac_ast)
        style = "b"
        if kid[1].name == "COLON":
            style = kid[2].token_text()
            kid = kid[2:]
        self.run_expression(kid[1])
        result = self.pop().value
        before = len(self.next_node_ids)
        if isinstance(result, Node):
            if style in ["b", "bfs"]:
                self.next_node_ids.add_obj(result, allow_dups=True)
            elif style in ["d", "dfs"]:
                self.next_node_ids.add_obj(result, push_front=True, allow_dups=True)
            else:
                self.rt_error(f"{style} is invalid take operation", kid[0])
        elif isinstance(result, JacSet):
            if style in ["b", "bfs"]:
                self.next_node_ids.add_obj_list(result, allow_dups=True)
            elif style in ["d", "dfs"]:
                self.next_node_ids.add_obj_list(
                    result, push_front=True, allow_dups=True
                )
            else:
                self.rt_error(f"{style} is invalid take operation", kid[0])
        elif result:
            self.rt_error(f"{result} is not destination type (i.e., nodes)", kid[1])
        after = len(self.next_node_ids)
        if before >= after and kid[2].name == "else_stmt":
            self.run_else_stmt(kid[2])
        after = len(self.next_node_ids)

    def run_disengage_action(self, jac_ast):
        """
        disengage_action: KW_DISENGAGE (report_action | SEMI);
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[1].name == "report_action":
            self.run_report_action(kid[1])
        self._stopped = "stop"
        self.next_node_ids.remove_all()

    def run_yield_action(self, jac_ast):
        """
        yield_action:
            KW_YIELD (
                report_action
                | disengage_action
                | take_action
                | SEMI
            );
        """
        kid = self.set_cur_ast(jac_ast)
        if len(kid) and kid[1].name != "SEMI":
            expr_func = getattr(self, f"run_{kid[1].name}")
            expr_func(kid[1])
        self.yield_walk()

    def run_preset_in_out(self, jac_ast, act):
        """
        preset_in_out:
            DBL_COLON param_list? (DBL_COLON | COLON_OUT expression);

        obj: The node or edge with preset
        act: The action associated with preset
        """
        kid = self.set_cur_ast(jac_ast)
        param_list = {"args": [], "kwargs": []}
        self.push_scope(
            JacScope(
                parent=self,
                name=f"p_in_out:{jac_ast.loc_str()}",
                has_obj=self.current_node,
                here=self.current_node,
                visitor=self,
            )
        )

        if kid[1].name == "param_list":
            param_list = self.run_param_list(kid[1]).value
        try:
            result = act.run_action(param_list, self._jac_scope, self)
        except Exception as e:
            self.rt_error(f"Internal Exception: {e}", self._cur_jac_ast)
            result = None
        if kid[-1].name == "expression":
            self.run_expression(kid[-1])
            dest = self.pop()
            dest.value = result
            dest.write(kid[-1])
        self.pop_scope()

    # Helper Functions ##################
    def auto_trigger_node_actions(self, act_list):
        already_executed = []  # handles inhereted duplicates, (overriding)
        nd = self.current_node
        for i in act_list.obj_list():
            if (
                i.access_list
                and self.name not in i.access_list
                or i.name in already_executed
            ):
                continue
            if i.preset_in_out:
                self.run_preset_in_out(jac_ir_to_ast(i.preset_in_out), i)
            else:
                self.call_ability(nd=nd, name=i.name, act_list=act_list)
            if not i.preset_in_out:  # All preset in and outs get executed
                already_executed.append(i.name)

    def scope_and_run(self, jac_ast, run_func, scope_name):
        """
        Helper to run ast elements with execution scope added
        (Useful for running arbitrary code blocks as one-offs)
        """
        self.push_scope(
            JacScope(
                parent=self,
                name=scope_name,
                has_obj=self,
                here=self.current_node,
                visitor=self,
            )
        )
        run_func(jac_ast)
        self.pop_scope()



"""Built in actions for Jaseci"""
import requests
from jaseci.jsorc.live_actions import jaseci_action
from bs4 import BeautifulSoup


@jaseci_action()
def get_page_meta(url: str, timeout: int = 3, parser: str = "lxml"):
    """
    Util to parse metadata out of urls and html documents
    Parser option: lxml (default), html5lib, html.parser
    Comparison between parsers: https://www.crummy.com/software/BeautifulSoup/bs4/doc/#installing-a-parser
    """
    try:
        webpage = requests.get(
            url,
            timeout=timeout,
            headers={
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:94.0) Gecko/20100101 Firefox/94.0"
            },
        )
        soup = BeautifulSoup(webpage.content, features=parser)
        meta = soup.find_all("meta")
        meta_list = []
        for tag in meta:
            meta_list.append(dict(tag.attrs))
        return meta_list
    except Exception as e:
        print("Failed")
        return f"Failed at getting metadata for {url}: {str(e)}"



"""
Admin config api functions as a mixin
"""

from typing import Optional
from json import dumps, loads
from jaseci.extens.api.interface import Interface


class ConfigApi:
    """
    Admin config APIs
    Abstracted since there are no valid configs in core atm, see jaseci_serv
    to see how used.
    """

    def __init__(self, *args, **kwargs):
        self._valid_configs = [
            "CONFIG_EXAMPLE",
            "ACTION_SETS",
            "REDIS_CONFIG",
            "TASK_CONFIG",
            "MAIL_CONFIG",
            "PROME_CONFIG",
            "ELASTIC_CONFIG",
            "STRIPE_CONFIG",
            "KUBE_CONFIG",
            "JSORC_CONFIG",
        ]

    @Interface.admin_api(cli_args=["name"])
    def config_get(self, name: str, do_check: bool = True):
        """
        Get a config
        """
        if do_check and not self.name_check(name):
            return self.name_error(name)
        return self._h.get_glob(name)

    @Interface.admin_api(cli_args=["name"])
    def config_set(self, name: str, value: str or dict, do_check: bool = True):
        """
        Set a config
        """
        if do_check and not self.name_check(name):
            return self.name_error(name)

        if not (value is None) and type(value) is dict:
            value = dumps(value)

        self._h.save_glob(name, value)
        return [f"Config of '{name}' to '{value}' set!"]

    @Interface.admin_api(cli_args=["name"])
    def config_update(
        self,
        name: str,
        field_key: str,
        field_value: str
        or int
        or float
        or list
        or dict
        or bool
        or tuple
        or None,  # json serializable types
        do_check: bool = True,
    ):
        """
        Update a key-value of a config
        """
        if do_check and not self.name_check(name):
            return self.name_error(name)

        conf = self._h.get_glob(name)
        try:
            conf = loads(conf)
        except Exception:
            return [
                f"Config {name} is not a dictionary. Uses config_set to set the value for this config."
            ]
        if field_key not in conf:
            return [f"Field {field_key} does not exist in config {name}"]

        conf[field_key] = field_value
        conf = dumps(conf)
        self._h.save_glob(name, conf)
        return [
            f"Config of '{name}' is updated with {field_key}:{field_value}. Current config value: {conf}"
        ]

    @Interface.admin_api()
    def config_list(self):
        """
        Check a config is present
        """
        return [v for v in self._h.list_glob() if v in self._valid_configs]

    @Interface.admin_api()
    def config_index(self):
        """
        List all valid configs
        """
        return self._valid_configs

    @Interface.admin_api(cli_args=["name"])
    def config_exists(self, name: str):
        """
        Check a config is present
        """
        return self._h.has_glob(name)

    @Interface.admin_api(cli_args=["name"])
    def config_delete(self, name: str, do_check: bool = True):
        """
        Delete a config
        """
        if do_check and not self.name_check(name):
            return self.name_error(name)
        self._h.destroy_glob(name)
        return [f"{name} Deleted."]

    def name_error(self, name):
        """Much used error output"""
        return [f"Config {name} not recognized, must be in {self._valid_configs}!"]

    def name_check(self, name):
        """Much used name check"""
        return name in self._valid_configs



"""
Variable scope manager for Jac

Utility for all runtime interaction with variables in different scopes
"""
from jaseci.jac.machine.jac_value import JacValue
from jaseci.jsorc.live_actions import get_global_actions


class JacScope:
    def __init__(self, parent, name, has_obj=None, here=None, visitor=None):
        self.parent = parent
        self.name = name
        self.local_scope = {}
        self.has_obj = has_obj if has_obj else self
        self.context = {}
        self.action_sets = []
        self._start_time = 0  # For profiling
        self._total_time = 0  # For profiling
        self._cum_start_time = None
        self.set_refs(here, visitor)
        self.setup_actions()

    def setup_actions(self):
        allactions = []
        for i in self.action_sets:
            allactions += i.obj_list()
        self.action_sets = get_global_actions()
        for i in allactions:
            self.add_action(i)

    def add_action(self, act):
        group = act.name.split(".")[0]
        if "." in act.name:
            if group not in self.action_sets.keys():
                self.action_sets[group] = {}
            action = act.name.split(".")[1]
            self.action_sets[group][action] = act
        else:
            self.action_sets[group] = act

    def set_refs(self, here, visitor):
        self.local_scope["here"] = here
        self.local_scope["visitor"] = visitor
        self.action_sets = []
        if here:
            self.action_sets += [here.get_architype().get_all_abilities()]
        if visitor:
            self.action_sets += [visitor.get_architype().get_all_abilities()]

    def get_refs(self):
        return {
            "here": self.local_scope["here"],
            "visitor": self.local_scope["visitor"],
        }

    def here(self):
        return self.local_scope["here"]

    def visitor(self):
        return self.local_scope["visitor"]

    def find_live_attr(self, name, allow_read_only=True):
        """Finds binding for variable if not in standard scope"""
        # check if var is in walker's context
        if "context" in self.has_obj.__dict__ and name in self.has_obj.context.keys():
            return JacValue(self.parent, ctx=self.has_obj, name=name)
        elif name in self.action_sets.keys():
            return JacValue(self.parent, ctx=self.action_sets, name=name)
        return None

    def get_live_var(self, name, create_mode=False):
        """Returns live variable"""
        found = None
        # Lock for variable in various locations
        if name in self.local_scope.keys():
            found = JacValue(self.parent, ctx=self.local_scope, name=name)
        else:
            found = self.find_live_attr(name)
        if found is None and create_mode:
            self.local_scope[name] = None
            return JacValue(self.parent, ctx=self.local_scope, name=name)
        if found:
            return found
        return None
